#ifndef TemporalInstance_H
#define TemporalInstance_H

//Bitsets are widely used in this implementation to encode the information of 
//	states, state expressions, and causal rules
#include <bitset>

#include <vector>
#include <map>
#include <string>
#include <fstream>

using namespace std;

//The maximum number of variables allowed in temporal reasoning
const int MAX_NUM_OF_VARIABLES = 16;


//*****************************************************************************************
// The State class
//*****************************************************************************************
class State
{
	private:
		// Internally, a state object uses a bitset to record the binary values of the state variables.
		// The balue of bit i records the value of state varibale X_i in this particular state   
		bitset<MAX_NUM_OF_VARIABLES> VariableValues;

	public:
		//Set the values of the  variables according toth egiven bitset
		void SetState(const bitset<MAX_NUM_OF_VARIABLES> & aBitset);

		//Set the values of the selected variables to 1; Return true if done successfully
		bool SetState(const vector<unsigned short> & vectOfVarIds);

		//Set values of all variables to 1
		void SetAll();

		//Reset the values of the selected variables to 0; Return true if done successfully
		bool ResetState(const vector<int> & vectOfVarIds);

		//Reset values of all variables to 0
		void ResetAll();

		//display the state
		void Display(const map<unsigned int, string> & variableIdToNameMap) const;

		//write the state information to a file
		void WriteToFile(const map<unsigned int, string> & variableIdToNameMap, ofstream & fout) const;

		//comparing two states
		inline bool operator <(const State  & anotherState) const
		{ return ( (this->VariableValues.to_ulong() ) < anotherState.VariableValues.to_ulong() ); };

		//comparing two states
		inline bool operator ==(const State  & anotherState) const
		{ return ( (this->VariableValues) == anotherState.VariableValues ); };

		//comparing two states
		inline bool operator !=(const State  & anotherState) const
		{ return ( (this->VariableValues) != anotherState.VariableValues ); };

		//Get the bitset representation of the state
		bitset<MAX_NUM_OF_VARIABLES>  GetVariableBitset() const;

	//The StateExpression class can access private members of State
	friend class StateExpression;

	//The CausalRule class can access private members of State
	friend class CausalRule;

	//TemporalInstance class can access private members of State
	friend class TemporalInstance;

};


//*****************************************************************************************
// The StateExpression class
//*****************************************************************************************		
class StateExpression
{ 
	private:
		//Information of the positive variables, which must be true in this state expression.
		//	The corresponding bits of theses variables are set to 1 in this bitset.
		bitset<MAX_NUM_OF_VARIABLES> PositiveConditions;


		//Information of the negative variables, which must be false in this state expression.
		//	The corresponding bits of theses variables are set to 1 in this bitset
		bitset<MAX_NUM_OF_VARIABLES> NegativeConditions;

	public:
		//Determine wehther the expression if true in a give state; return true if so
		bool IsTrueIn(const State & thisState) const;

		//Set the two variable bitsets according to the information of the IDs of
		//	the positive or negative variables respectively.
		bool SetStateExpression
		(	const vector<unsigned short> & vectOfVarIdsOfPositiveConditions, 
			const vector<unsigned short> & vectOfVarIdsOfNegativeConditions	);

		//display a state expression with respect to a map from variable IDs to names
		void Display(const map<unsigned int, string> & VariableIdToNameMap) const;

		//write to file a state expression with respect to a map from variable IDs to names
		void WriteToFile(const map<unsigned int, string> & VariableIdToNameMap, ofstream & fout) const;

	//The CausalRule class can access private members of StateExpression
	friend class CausalRule;

	//TemporalInstance class can access private members of StateExpression
	friend class TemporalInstance;


};


//*****************************************************************************************
//The CausalRule class
//*****************************************************************************************
class CausalRule
{ 
	private:
		//The preconditions represented as a StateExpression
		StateExpression PreConditions;

		//The postconditions represented as a StateExpression
		StateExpression PostConditions;

	public:
		//Get the next state from thisState 
		State GetNextStateFrom(const State & thisState) const;

		//Set up the causal rule representation; Return true if successfully done
		bool SetCausalRule
		(	const vector<unsigned short> & vectOfVarIdsOfPositivePreConditions, 
			const vector<unsigned short> & vectOfVarIdsOfNegativePreConditions,	
			const vector<unsigned short> & vectOfVarIdsOfPositivePostConditions, 
			const vector<unsigned short> & vectOfVarIdsOfNegativePostConditions	
		);

		//display a causal rule with respect to a map from variable IDs to names
		void Display(const map<unsigned int, string> & VariableIdToNameMap) const;

		//write to a file a causal rule with respect to a map from variable IDs to names
		void WriteToFile
			(const map<unsigned int, string> & VariableIdToNameMap, ofstream & fout) const;

		//TemporalInstance class can access private members of CausalRules
	friend class TemporalInstance;
};


//*****************************************************************************************
// The Event class
//*****************************************************************************************


class Event
{
	public:
		//Define double as the type for point-of-time information
		typedef double TimePoint;

	private:
		//Name of the event
		string EventName;

		//The causal rules
		vector<CausalRule> CausalRules;

		//Must occur after this point of time
		TimePoint OccurAfterThisTime;

		//Must occur before this point of time
		TimePoint OccurBeforeThisTime;

	public:

		//display an event with respect to a map from variable IDs to names
		void Display(map<unsigned int, string> & VariableIdToNameMap);

		//Add a new causal rule
		void AddCausalRule(CausalRule & aRule);

		//Set the name
		void SetName(string aName);

		//Set the two time points OccurAfterThisTime and OccurBeforeThisTime
		void SetTime(TimePoint OA, TimePoint OB);

		//Determine whether one event can only strictly occur after another
		bool operator < (Event & anoterEvent);

		//Determine whether one event can only strictly occur before another
		bool operator > (Event & anoterEvent);

		//Determine whether the time intervals of two events overlap
		bool Overlap(Event & anoterEvent);

	//TemporalInstance class can access private members of events
	friend class TemporalInstance;

	//TemporalSolver class can access private members of events
	friend class TemporalSolver;

	//Compare_Events_By_Their_EarliestStartingPoints class can access private members of events
	friend class Compare_Events_By_Their_EarliestStartingPoints;
};


//A class to provide functional objects for comparing
//	events based on their earliest possible starting time
class Compare_Events_By_Their_EarliestStartingPoints
{public:
	bool operator()(Event eventA, Event eventB)
	{ return (eventA.OccurAfterThisTime < eventB.OccurAfterThisTime);
	}
};


//*****************************************************************************************
//The TemporalInstance class 
//*****************************************************************************************
class TemporalInstance
{
	private:
		//Instance name
		string InstanceName;

		//Initial state
		State InitialState;

		//Goal
		StateExpression Goal;

		//A map for mapping the names of the events into their IDs
		map<string, unsigned short> VariableNameToIdMap;

		//A map for mapping the IDs of the events into their names
		map<unsigned int, string> VariableIdToNameMap;

		//The events
		vector<Event> Events;

		//Read the lines of description from a file into a vector of strings
		//	a privatemethod called by 
		bool ReadLinesFromFile(char * fileName, vector<string> &linesOfDescription);

	public:

		//Check whether the time interval information is consistent
		bool IsTimeInformationConsistent() const;

		//Read the contents of a temporal instance from a file
		bool ReadInstanceFromFile(char * fileName);

		//Display the contents of a temporal instance
		void Display();


	//The TemporalSolver class is allowed to acccess the prvate members of TemporalInstance
	friend class TemporalSolver;

};


	
//*****************************************************************************************
//*****************************************************************************************
#endif
