#ifndef temporalProjection_H
#define temporalProjection_H

//Bitsets are widely used in this implementation to encode the information of 
//	states, state expressions, and causal rules
#include "temporalInstance.h"
#include <vector>

using namespace std;

//*****************************************************************************************
// The TemporalSolver class
//*****************************************************************************************
class TemporalSolver
{
	public:
		//A class defining the search nodes used in this temporal projection implementation
		class SearchNode
		{public:
			//The system state 
			State SystemState;

			//Information of the numbers of events already occurred in individual chains
			vector<unsigned int> ProgressInChains;
		};

		//A class defining the trace labels used in this temporal projection implementation
		//	to trace the search nodes and the trajectory of an event sequence
		class TraceLabel
		{public:
			//The previous search node 
			SearchNode PreviousSearchNode;

			//Information of the event and the associated rule that triggers 
			//	the state transition in that search node
			unsigned int EventId;
			unsigned short RuleId;
		};

		//A class defining the format to store the Solution of temporal projection
		class Solution
		{public:
			const TemporalInstance * PtrToTemporalInstance;
			unsigned int SpecialEventId;
			Event::TimePoint SpecialTimePoint;

			bool IsGoalReachable;
			vector< TraceLabel > EventTrajectory;
			vector< Event::TimePoint > EventTimeLine;
		};

	private:
		Solution mySolution;

	public:
		//A static class constant to indicate no particular event is of concern
		static const unsigned int NO_PARTICULAR_EVENT;

		//A static class constant to indicate no particular time point is of concern
		static const Event::TimePoint NO_PARTICULAR_TIMEPOINT;

		//This method is called to Calculate the event chains
		void GetEventChains(	const TemporalInstance & myInstance, 
								vector< vector<unsigned int> > & eventIdChains);

		//This method is called to display the informtion of the event chains
		void DisplayEventChains(	const TemporalInstance & myInstance, 
								vector< vector<unsigned int> > & eventIdChains);

		//This method is called to display the informtion of a search node
		void DisplaySearchNode(	const TemporalInstance & myInstance, const SearchNode & node);

		//This method is called to get a safe time unit to seperate events in a timeline
		Event::TimePoint GetTimelineDelta(	const TemporalInstance & myInstance);

		//Determine what might be the next chain to progress
		void GetAllPossibleNextChainsToProgress
			(	const TemporalInstance & myInstance, 
				const vector< vector<unsigned int> > & eventIdChains,
				const vector<unsigned int> & ProgressInChains,
				vector< unsigned int > & possibleNextChainsToProgress);

		//Try to find a solution and store the Solution in the repositor.
		//	return true if there is a solution to reach the goal
		bool GetTemporalSolverSolution(const TemporalInstance & myInstance, Solution & mySolution);

		//Display a solution 
		void DisplaySolution(const Solution& mySolution) const;

		//Write a solution to a file
		void WriteSolutionToFile(const Solution& mySolution, const char * fileName) const;


};
	
//*****************************************************************************************
//*****************************************************************************************
#endif
